# File:         builder.py
# Author:       Brian Allen Vanderburg II
# Purpose:      The builder class builds the output from xml files
# License:      Refer to the file license.txt

import os
import re

from lxml import etree

from . import util


class Builder(object):
    def __init__(self, xml):
        # Basic stuff
        self.extension = xml.get('extension', '.html')
        self.encoding = xml.get('encoding', 'utf-8')
        self.doctype = xml.get('doctype', '<!DOCTYPE html>')
        self.strip = util.getbool(xml.get('strip', 'no'))

        # Parameters
        self.params = {}
        for i in xml.findall('param'):
            name = i.get('name')
            value = i.get('value')
            if name and value:
                self.params[name] = value;

        # Transforms
        self.transforms = {}
        for i in xml.findall('transform'):
            self.transforms[i.get('root')] = i.get('xsl')

        # Header and footer
        header = xml.find('header');
        self.header = ''
        if header is not None:
            self.header = header.text;
        
        footer = xml.find('footer');
        self.footer = ''
        if footer is not None:
            self.footer = footer.text;

        # Fix closing tags
        self.closetags = []
        for i in xml.findall('closetag'):
            self.closetags.append(i.get('tag'))

        # Preserve tags during HTML reduction
        self.preservetags = []
        for i in xml.findall('preservetag'):
            self.preservetags.append(i.get('tag'))

        # Find and replace
        self.replacements = []
        for i in xml.findall('find'):
            self.replacements.append((i.get('match'), i.get('replace')))
    
    @staticmethod
    def load(xml):
        return Builder(xml)

    def execute(self, profile, params, source, target, relpath, ending):
        util.message('Transforming: ' + relpath)

        sourcefile = os.path.join(source, relpath)
        reldest = relpath[:-len(ending)] + self.extension
        targetfile = os.path.join(target, reldest)

        # Is the page out of date?
        if os.path.isfile(targetfile):
            stime = os.stat(sourcefile).st_mtime
            ttime = os.stat(targetfile).st_mtime

            if stime < ttime:
                util.status('NC')
                return

            os.unlink(targetfile)

        # Parameters
        sourcedir = os.path.dirname(sourcefile)
        targetdir = os.path.dirname(targetfile)

        coreparams = {
            'sourceroot': source.replace(os.sep, '/').rstrip('/') + '/',
            'targetroot': target.replace(os.sep, '/').rstrip('/') + '/',
            'sourcedir': sourcedir.replace(os.sep, '/').rstrip('/') + '/',
            'targetdir': targetdir.replace(os.sep, '/').rstrip('/') + '/',
            'sourcefile': sourcefile.replace(os.sep, '/'),
            'targetfile': targetfile.replace(os.sep, '/'),
            'sourcerpath': relpath.replace(os.sep, '/'),
            'targetrpath': reldest.replace(os.sep, '/'),
            'relativeroot':  '../' * relpath.count(os.sep),
            'profile': profile
        }

        bparams = self.params.copy()
        bparams.update(params)
        bparams.update(coreparams)

        # Build
        if self.build(sourcefile, targetfile, bparams):
            util.status('OK')
        else:
            util.status('IGN')

    def build(self, sourcefile, targetfile, params):
        inxml = etree.parse(sourcefile);
        inxml.xinclude()
        out = self.buildhtml(inxml, params)

        if not out is None:
            targetdir = os.path.dirname(targetfile)

            if not os.path.isdir(targetdir):
                os.makedirs(targetdir)

            file(targetfile, 'wb').write(out.encode(self.encoding))
            return True
        else:
            return False

    def cleanup(self, output, params):
        # Remove <?xml .. ?>
        if output[:2] == '<?':
            pos = output.find('?>')
            if pos >= 0:
                output = output[pos + 2:]
                output = output.lstrip()

        # Replace <!DOCTYPE ... >
        if len(self.doctype) > 0:
            if output[:2] == '<!':
                pos = output.find('>')
                if pos > 0:
                    output = self.doctype + output[pos + 1:]
            else:
                output = self.doctype + '\n' + output

        # Fix closing tags: <tag /> -> <tag></tag>
        if len(self.closetags) > 0:
            output = re.sub(r'(?si)<(' + r'|'.join(self.closetags) + r')(((\s[^>]*?)?)/>)', r'<\1\3></\1>', output)

        # <tag/> -> <tag />
        output = re.sub(r'(?si)<(([a-zA-Z0-9]|:)+)([^>]*?)/>', r'<\1\3 />', output)

        # Find and replace
        for pair in self.replacements:
            output = output.replace(pair[0], pair[1])

        # Strip whitespace
        if self.strip:
            pos = 0
            result = ''

            # Make sure to preserve certain tags
            if len(self.preservetags) > 0:
                pattern = r'(?si)<(' + r'|'.join(self.preservetags) + r')(>|\s[^>]*?>).*?</\1>'
                matches = re.finditer(pattern, output)

                for match in matches:
                    offset = match.start()
                    if offset > pos:
                        result += re.sub(r'(?m)^\s+', r'', output[pos:offset])
                
                    result += match.group()
                    pos = match.end()

            # Any leftover is also stripped
            if pos < len(output):
                result += re.sub(r'(?m)^\s+', r'', output[pos:])

            output = result
    
        # Add header and footer to output
        def helper(mo):
            key = mo.group(1)
            if key == '':
                return '@';
            return params[key]

        if len(self.header) > 0:
            output = re.sub('@([a-zA-Z0-9]*?)@', helper, self.header) + output

        if len(self.footer) > 0:
            output = output + re.sub('@([a-zA-Z0-9]*?)@', helper, self.footer)

        return output.strip()


    _cache = {}
    @classmethod
    def getxslt(cls, path):
        if not path in cls._cache:
            xslxml = etree.parse(path)
            xslxml.xinclude()
            cls._cache[path] = etree.XSLT(xslxml)

        return cls._cache[path]

    def buildxml(self, xml, params):
        from .config import Config

        # Prepare parameters
        params = dict(params)
        for i in params:
            params[i] = etree.XSLT.strparam(params[i])

        # Build
        result = None
        root = xml.getroot()
        if root.tag in self.transforms:
            transform = self.getxslt(Config.path(self.transforms[root.tag]))
            result = transform(xml, **params)

        return result

    def buildhtml(self, xml, params):
        result = self.buildxml(xml, params)
        if not result is None:
            result = self.cleanup(etree.tostring(result, pretty_print=True), params)

        return result

